@@ -5,7 +5,7 @@
  * The contents of this file are subject to the Interbase Public
  * License Version 1.0 (the "License"); you may not use this file
  * except in compliance with the License. You may obtain a copy
- * of the License at http://www.firebirdsql.org/index.php?op=doc&id=ipl
+ * of the License at http://www.Inprise.com/IPL.html
  *
  * Software distributed under the License is distributed on an
  * "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
@@ -31,6 +31,8 @@
 /**********************************/
 
 #define isc_dpb_version1                  1
+#define isc_dpb_version2                  2
+
 #define isc_dpb_cdd_pathname              1
 #define isc_dpb_allocation                2
 #define isc_dpb_journal                   3
@@ -99,7 +101,7 @@
 #define isc_dpb_gfix_attach               66
 #define isc_dpb_gstat_attach              67
 #define isc_dpb_set_db_charset            68
-#define isc_dpb_gsec_attach               69
+#define isc_dpb_gsec_attach               69		/* deprecated */
 #define isc_dpb_address_path              70
 #define isc_dpb_process_id                71
 #define isc_dpb_no_db_triggers            72
@@ -109,6 +111,18 @@
 #define isc_dpb_org_filename			  76
 #define isc_dpb_utf8_filename			  77
 #define isc_dpb_ext_call_depth			  78
+#define isc_dpb_auth_block				  79
+#define isc_dpb_client_version			  80
+#define isc_dpb_remote_protocol			  81
+#define isc_dpb_host_name				  82
+#define isc_dpb_os_user					  83
+#define isc_dpb_specific_auth_data		  84
+#define isc_dpb_auth_plugin_list		  85
+#define isc_dpb_auth_plugin_name		  86
+#define isc_dpb_config					  87
+#define isc_dpb_nolinger				  88
+#define isc_dpb_reset_icu				  89
+#define isc_dpb_map_attach                90
 
 /**************************************************/
 /* clumplet tags used inside isc_dpb_address_path */
@@ -246,6 +260,7 @@
 #define isc_spb_version1                  1
 #define isc_spb_current_version           2
 #define isc_spb_version                   isc_spb_current_version
+#define isc_spb_version3                  3
 #define isc_spb_user_name                 isc_dpb_user_name
 #define isc_spb_sys_user_name             isc_dpb_sys_user_name
 #define isc_spb_sys_user_name_enc         isc_dpb_sys_user_name_enc
@@ -260,12 +275,25 @@
 #define isc_spb_trusted_auth			  111
 #define isc_spb_process_name              112
 #define isc_spb_trusted_role              113
-
+#define isc_spb_verbint                   114
+#define isc_spb_auth_block                115
+#define isc_spb_auth_plugin_name          116
+#define isc_spb_auth_plugin_list          117
+#define isc_spb_utf8_filename			  118
+#define isc_spb_client_version            119
+#define isc_spb_remote_protocol           120
+#define isc_spb_host_name                 121
+#define isc_spb_os_user                   122
+#define isc_spb_config					  123
+#define isc_spb_expected_db				  124
 
 #define isc_spb_connect_timeout           isc_dpb_connect_timeout
 #define isc_spb_dummy_packet_interval     isc_dpb_dummy_packet_interval
 #define isc_spb_sql_role_name             isc_dpb_sql_role_name
 
+// This will not be used in protocol 13, therefore may be reused
+#define isc_spb_specific_auth_data		  isc_spb_trusted_auth
+
 /*****************************
  * Service action items      *
  *****************************/
@@ -293,32 +321,35 @@
 #define isc_action_svc_set_mapping    27	// Set auto admins mapping in security database
 #define isc_action_svc_drop_mapping   28	// Drop auto admins mapping in security database
 #define isc_action_svc_display_user_adm 29	// Displays user(s) from security database with admin info
-#define isc_action_svc_last			  30	// keep it last !
+#define isc_action_svc_validate		  30	// Starts database online validation
+#define isc_action_svc_last			  31	// keep it last !
 
 /*****************************
  * Service information items *
  *****************************/
 
-#define isc_info_svc_svr_db_info      50	/* Retrieves the number of attachments and databases */
-#define isc_info_svc_get_license      51	/* Retrieves all license keys and IDs from the license file */
-#define isc_info_svc_get_license_mask 52	/* Retrieves a bitmask representing licensed options on the server */
-#define isc_info_svc_get_config       53	/* Retrieves the parameters and values for IB_CONFIG */
-#define isc_info_svc_version          54	/* Retrieves the version of the services manager */
-#define isc_info_svc_server_version   55	/* Retrieves the version of the Firebird server */
-#define isc_info_svc_implementation   56	/* Retrieves the implementation of the Firebird server */
-#define isc_info_svc_capabilities     57	/* Retrieves a bitmask representing the server's capabilities */
-#define isc_info_svc_user_dbpath      58	/* Retrieves the path to the security database in use by the server */
-#define isc_info_svc_get_env	      59	/* Retrieves the setting of $FIREBIRD */
-#define isc_info_svc_get_env_lock     60	/* Retrieves the setting of $FIREBIRD_LCK */
-#define isc_info_svc_get_env_msg      61	/* Retrieves the setting of $FIREBIRD_MSG */
-#define isc_info_svc_line             62	/* Retrieves 1 line of service output per call */
-#define isc_info_svc_to_eof           63	/* Retrieves as much of the server output as will fit in the supplied buffer */
-#define isc_info_svc_timeout          64	/* Sets / signifies a timeout value for reading service information */
-#define isc_info_svc_get_licensed_users 65	/* Retrieves the number of users licensed for accessing the server */
-#define isc_info_svc_limbo_trans	66	/* Retrieve the limbo transactions */
-#define isc_info_svc_running		67	/* Checks to see if a service is running on an attachment */
-#define isc_info_svc_get_users		68	/* Returns the user information from isc_action_svc_display_users */
-#define isc_info_svc_stdin			78	/* Returns size of data, needed as stdin for service */
+#define isc_info_svc_svr_db_info		50	/* Retrieves the number of attachments and databases */
+#define isc_info_svc_get_license		51	/* Retrieves all license keys and IDs from the license file */
+#define isc_info_svc_get_license_mask	52	/* Retrieves a bitmask representing licensed options on the server */
+#define isc_info_svc_get_config			53	/* Retrieves the parameters and values for IB_CONFIG */
+#define isc_info_svc_version			54	/* Retrieves the version of the services manager */
+#define isc_info_svc_server_version		55	/* Retrieves the version of the Firebird server */
+#define isc_info_svc_implementation		56	/* Retrieves the implementation of the Firebird server */
+#define isc_info_svc_capabilities		57	/* Retrieves a bitmask representing the server's capabilities */
+#define isc_info_svc_user_dbpath		58	/* Retrieves the path to the security database in use by the server */
+#define isc_info_svc_get_env			59	/* Retrieves the setting of $FIREBIRD */
+#define isc_info_svc_get_env_lock		60	/* Retrieves the setting of $FIREBIRD_LCK */
+#define isc_info_svc_get_env_msg		61	/* Retrieves the setting of $FIREBIRD_MSG */
+#define isc_info_svc_line				62	/* Retrieves 1 line of service output per call */
+#define isc_info_svc_to_eof				63	/* Retrieves as much of the server output as will fit in the supplied buffer */
+#define isc_info_svc_timeout			64	/* Sets / signifies a timeout value for reading service information */
+#define isc_info_svc_get_licensed_users	65	/* Retrieves the number of users licensed for accessing the server */
+#define isc_info_svc_limbo_trans		66	/* Retrieve the limbo transactions */
+#define isc_info_svc_running			67	/* Checks to see if a service is running on an attachment */
+#define isc_info_svc_get_users			68	/* Returns the user information from isc_action_svc_display_users */
+#define isc_info_svc_auth_block			69	/* Sets authentication block for service query() call */
+#define isc_info_svc_stdin				78	/* Returns maximum size of data, needed as stdin for service */
+
 
 /******************************************************
  * Parameters for isc_action_{add|del|mod|disp)_user  *
@@ -351,6 +382,8 @@
 #define isc_spb_bkp_file                 5
 #define isc_spb_bkp_factor               6
 #define isc_spb_bkp_length               7
+#define isc_spb_bkp_skip_data            8
+#define isc_spb_bkp_stat                 15
 #define isc_spb_bkp_ignore_checksums     0x01
 #define isc_spb_bkp_ignore_limbo         0x02
 #define isc_spb_bkp_metadata_only        0x04
@@ -376,6 +409,7 @@
 #define isc_spb_prp_set_sql_dialect		14
 #define isc_spb_prp_activate			0x0100
 #define isc_spb_prp_db_online			0x0200
+#define isc_spb_prp_nolinger			0x0400
 #define isc_spb_prp_force_shutdown			41
 #define isc_spb_prp_attachments_shutdown	42
 #define isc_spb_prp_transactions_shutdown	43
@@ -434,6 +468,12 @@
 #define isc_spb_tra_advise_commit		30
 #define isc_spb_tra_advise_rollback		31
 #define isc_spb_tra_advise_unknown		33
+#define isc_spb_tra_id_64				46
+#define isc_spb_single_tra_id_64		47
+#define isc_spb_multi_tra_id_64			48
+#define isc_spb_rpr_commit_trans_64		49
+#define isc_spb_rpr_rollback_trans_64	50
+#define isc_spb_rpr_recover_two_phase_64	51
 
 #define isc_spb_rpr_validate_db			0x01
 #define isc_spb_rpr_sweep_db			0x02
@@ -443,17 +483,20 @@
 #define isc_spb_rpr_ignore_checksum		0x20
 #define isc_spb_rpr_kill_shadows		0x40
 #define isc_spb_rpr_full				0x80
+#define isc_spb_rpr_icu				  0x0800
 
 /*****************************************
  * Parameters for isc_action_svc_restore *
  *****************************************/
 
+#define isc_spb_res_skip_data			isc_spb_bkp_skip_data
 #define isc_spb_res_buffers				9
 #define isc_spb_res_page_size			10
 #define isc_spb_res_length				11
 #define isc_spb_res_access_mode			12
 #define isc_spb_res_fix_fss_data		13
 #define isc_spb_res_fix_fss_metadata	14
+#define isc_spb_res_stat				isc_spb_bkp_stat
 #define isc_spb_res_metadata_only		isc_spb_bkp_metadata_only
 #define isc_spb_res_deactivate_idx		0x0100
 #define isc_spb_res_no_shadow			0x0200
@@ -463,6 +506,16 @@
 #define isc_spb_res_create				0x2000
 #define isc_spb_res_use_all_space		0x4000
 
+/*****************************************
+ * Parameters for isc_action_svc_validate *
+ *****************************************/
+
+#define isc_spb_val_tab_incl		1	// include filter based on regular expression
+#define isc_spb_val_tab_excl		2	// exclude filter based on regular expression
+#define isc_spb_val_idx_incl		3	// regexp of indices to validate
+#define isc_spb_val_idx_excl		4	// regexp of indices to NOT validate
+#define isc_spb_val_lock_timeout	5	// how long to wait for table lock
+
 /******************************************
  * Parameters for isc_spb_res_access_mode  *
  ******************************************/
@@ -481,14 +534,17 @@
  * Parameters for isc_info_svc_db_stats  *
  *****************************************/
 
+#define isc_spb_sts_table			64
+
 #define isc_spb_sts_data_pages		0x01
 #define isc_spb_sts_db_log			0x02
 #define isc_spb_sts_hdr_pages		0x04
 #define isc_spb_sts_idx_pages		0x08
 #define isc_spb_sts_sys_relations	0x10
 #define isc_spb_sts_record_versions	0x20
-#define isc_spb_sts_table			0x40
+//#define isc_spb_sts_table			0x40
 #define isc_spb_sts_nocreation		0x80
+#define isc_spb_sts_encryption	   0x100
 
 /***********************************/
 /* Server configuration key values */
@@ -513,393 +569,6 @@
 #define isc_spb_trc_name			2
 #define isc_spb_trc_cfg				3
 
-/**********************************************/
-/* Dynamic Data Definition Language operators */
-/**********************************************/
-
-/******************/
-/* Version number */
-/******************/
-
-#define isc_dyn_version_1                 1
-#define isc_dyn_eoc                       255
-
-/******************************/
-/* Operations (may be nested) */
-/******************************/
-
-#define isc_dyn_begin                     2
-#define isc_dyn_end                       3
-#define isc_dyn_if                        4
-#define isc_dyn_def_database              5
-#define isc_dyn_def_global_fld            6
-#define isc_dyn_def_local_fld             7
-#define isc_dyn_def_idx                   8
-#define isc_dyn_def_rel                   9
-#define isc_dyn_def_sql_fld               10
-#define isc_dyn_def_view                  12
-#define isc_dyn_def_trigger               15
-#define isc_dyn_def_security_class        120
-#define isc_dyn_def_dimension             140
-#define isc_dyn_def_generator             24
-#define isc_dyn_def_function              25
-#define isc_dyn_def_filter                26
-#define isc_dyn_def_function_arg          27
-#define isc_dyn_def_shadow                34
-#define isc_dyn_def_trigger_msg           17
-#define isc_dyn_def_file                  36
-#define isc_dyn_mod_database              39
-#define isc_dyn_mod_rel                   11
-#define isc_dyn_mod_global_fld            13
-#define isc_dyn_mod_idx                   102
-#define isc_dyn_mod_local_fld             14
-#define isc_dyn_mod_sql_fld		  216
-#define isc_dyn_mod_view                  16
-#define isc_dyn_mod_security_class        122
-#define isc_dyn_mod_trigger               113
-#define isc_dyn_mod_trigger_msg           28
-#define isc_dyn_delete_database           18
-#define isc_dyn_delete_rel                19
-#define isc_dyn_delete_global_fld         20
-#define isc_dyn_delete_local_fld          21
-#define isc_dyn_delete_idx                22
-#define isc_dyn_delete_security_class     123
-#define isc_dyn_delete_dimensions         143
-#define isc_dyn_delete_trigger            23
-#define isc_dyn_delete_trigger_msg        29
-#define isc_dyn_delete_filter             32
-#define isc_dyn_delete_function           33
-#define isc_dyn_delete_shadow             35
-#define isc_dyn_grant                     30
-#define isc_dyn_revoke                    31
-#define isc_dyn_revoke_all                246
-#define isc_dyn_def_primary_key           37
-#define isc_dyn_def_foreign_key           38
-#define isc_dyn_def_unique                40
-#define isc_dyn_def_procedure             164
-#define isc_dyn_delete_procedure          165
-#define isc_dyn_def_parameter             135
-#define isc_dyn_delete_parameter          136
-#define isc_dyn_mod_procedure             175
-/* Deprecated.
-#define isc_dyn_def_log_file              176
-#define isc_dyn_def_cache_file            180
-*/
-#define isc_dyn_def_exception             181
-#define isc_dyn_mod_exception             182
-#define isc_dyn_del_exception             183
-/* Deprecated.
-#define isc_dyn_drop_log                  194
-#define isc_dyn_drop_cache                195
-#define isc_dyn_def_default_log           202
-*/
-#define isc_dyn_def_difference            220
-#define isc_dyn_drop_difference           221
-#define isc_dyn_begin_backup              222
-#define isc_dyn_end_backup                223
-#define isc_dyn_debug_info                240
-
-/***********************/
-/* View specific stuff */
-/***********************/
-
-#define isc_dyn_view_blr                  43
-#define isc_dyn_view_source               44
-#define isc_dyn_view_relation             45
-#define isc_dyn_view_context              46
-#define isc_dyn_view_context_name         47
-
-/**********************/
-/* Generic attributes */
-/**********************/
-
-#define isc_dyn_rel_name                  50
-#define isc_dyn_fld_name                  51
-#define isc_dyn_new_fld_name              215
-#define isc_dyn_idx_name                  52
-#define isc_dyn_description               53
-#define isc_dyn_security_class            54
-#define isc_dyn_system_flag               55
-#define isc_dyn_update_flag               56
-#define isc_dyn_prc_name                  166
-#define isc_dyn_prm_name                  137
-#define isc_dyn_sql_object                196
-#define isc_dyn_fld_character_set_name    174
-
-/********************************/
-/* Relation specific attributes */
-/********************************/
-
-#define isc_dyn_rel_dbkey_length          61
-#define isc_dyn_rel_store_trig            62
-#define isc_dyn_rel_modify_trig           63
-#define isc_dyn_rel_erase_trig            64
-#define isc_dyn_rel_store_trig_source     65
-#define isc_dyn_rel_modify_trig_source    66
-#define isc_dyn_rel_erase_trig_source     67
-#define isc_dyn_rel_ext_file              68
-#define isc_dyn_rel_sql_protection        69
-#define isc_dyn_rel_constraint            162
-#define isc_dyn_delete_rel_constraint     163
-
-#define isc_dyn_rel_temporary				238
-#define isc_dyn_rel_temp_global_preserve	1
-#define isc_dyn_rel_temp_global_delete		2
-
-/************************************/
-/* Global field specific attributes */
-/************************************/
-
-#define isc_dyn_fld_type                  70
-#define isc_dyn_fld_length                71
-#define isc_dyn_fld_scale                 72
-#define isc_dyn_fld_sub_type              73
-#define isc_dyn_fld_segment_length        74
-#define isc_dyn_fld_query_header          75
-#define isc_dyn_fld_edit_string           76
-#define isc_dyn_fld_validation_blr        77
-#define isc_dyn_fld_validation_source     78
-#define isc_dyn_fld_computed_blr          79
-#define isc_dyn_fld_computed_source       80
-#define isc_dyn_fld_missing_value         81
-#define isc_dyn_fld_default_value         82
-#define isc_dyn_fld_query_name            83
-#define isc_dyn_fld_dimensions            84
-#define isc_dyn_fld_not_null              85
-#define isc_dyn_fld_precision             86
-#define isc_dyn_fld_char_length           172
-#define isc_dyn_fld_collation             173
-#define isc_dyn_fld_default_source        193
-#define isc_dyn_del_default               197
-#define isc_dyn_del_validation            198
-#define isc_dyn_single_validation         199
-#define isc_dyn_fld_character_set         203
-#define isc_dyn_del_computed              242
-
-/***********************************/
-/* Local field specific attributes */
-/***********************************/
-
-#define isc_dyn_fld_source                90
-#define isc_dyn_fld_base_fld              91
-#define isc_dyn_fld_position              92
-#define isc_dyn_fld_update_flag           93
-
-/*****************************/
-/* Index specific attributes */
-/*****************************/
-
-#define isc_dyn_idx_unique                100
-#define isc_dyn_idx_inactive              101
-#define isc_dyn_idx_type                  103
-#define isc_dyn_idx_foreign_key           104
-#define isc_dyn_idx_ref_column            105
-#define isc_dyn_idx_statistic             204
-
-/*******************************/
-/* Trigger specific attributes */
-/*******************************/
-
-#define isc_dyn_trg_type                  110
-#define isc_dyn_trg_blr                   111
-#define isc_dyn_trg_source                112
-#define isc_dyn_trg_name                  114
-#define isc_dyn_trg_sequence              115
-#define isc_dyn_trg_inactive              116
-#define isc_dyn_trg_msg_number            117
-#define isc_dyn_trg_msg                   118
-
-/**************************************/
-/* Security Class specific attributes */
-/**************************************/
-
-#define isc_dyn_scl_acl                   121
-#define isc_dyn_grant_user                130
-#define isc_dyn_grant_user_explicit       219
-#define isc_dyn_grant_proc                186
-#define isc_dyn_grant_trig                187
-#define isc_dyn_grant_view                188
-#define isc_dyn_grant_options             132
-#define isc_dyn_grant_user_group          205
-#define isc_dyn_grant_role                218
-#define isc_dyn_grant_grantor			  245
-
-
-/**********************************/
-/* Dimension specific information */
-/**********************************/
-
-#define isc_dyn_dim_lower                 141
-#define isc_dyn_dim_upper                 142
-
-/****************************/
-/* File specific attributes */
-/****************************/
-
-#define isc_dyn_file_name                 125
-#define isc_dyn_file_start                126
-#define isc_dyn_file_length               127
-#define isc_dyn_shadow_number             128
-#define isc_dyn_shadow_man_auto           129
-#define isc_dyn_shadow_conditional        130
-
-/********************************/
-/* Log file specific attributes */
-/********************************/
-/* Deprecated.
-#define isc_dyn_log_file_sequence         177
-#define isc_dyn_log_file_partitions       178
-#define isc_dyn_log_file_serial           179
-#define isc_dyn_log_file_overflow         200
-#define isc_dyn_log_file_raw              201
-*/
-
-/***************************/
-/* Log specific attributes */
-/***************************/
-/* Deprecated.
-#define isc_dyn_log_group_commit_wait     189
-#define isc_dyn_log_buffer_size           190
-#define isc_dyn_log_check_point_length    191
-#define isc_dyn_log_num_of_buffers        192
-*/
-
-/********************************/
-/* Function specific attributes */
-/********************************/
-
-#define isc_dyn_function_name             145
-#define isc_dyn_function_type             146
-#define isc_dyn_func_module_name          147
-#define isc_dyn_func_entry_point          148
-#define isc_dyn_func_return_argument      149
-#define isc_dyn_func_arg_position         150
-#define isc_dyn_func_mechanism            151
-#define isc_dyn_filter_in_subtype         152
-#define isc_dyn_filter_out_subtype        153
-
-
-#define isc_dyn_description2              154
-#define isc_dyn_fld_computed_source2      155
-#define isc_dyn_fld_edit_string2          156
-#define isc_dyn_fld_query_header2         157
-#define isc_dyn_fld_validation_source2    158
-#define isc_dyn_trg_msg2                  159
-#define isc_dyn_trg_source2               160
-#define isc_dyn_view_source2              161
-#define isc_dyn_xcp_msg2                  184
-
-/*********************************/
-/* Generator specific attributes */
-/*********************************/
-
-#define isc_dyn_generator_name            95
-#define isc_dyn_generator_id              96
-
-/*********************************/
-/* Procedure specific attributes */
-/*********************************/
-
-#define isc_dyn_prc_inputs                167
-#define isc_dyn_prc_outputs               168
-#define isc_dyn_prc_source                169
-#define isc_dyn_prc_blr                   170
-#define isc_dyn_prc_source2               171
-#define isc_dyn_prc_type                  239
-
-#define isc_dyn_prc_t_selectable          1
-#define isc_dyn_prc_t_executable          2
-
-/*********************************/
-/* Parameter specific attributes */
-/*********************************/
-
-#define isc_dyn_prm_number                138
-#define isc_dyn_prm_type                  139
-#define isc_dyn_prm_mechanism             241
-
-/********************************/
-/* Relation specific attributes */
-/********************************/
-
-#define isc_dyn_xcp_msg                   185
-
-/**********************************************/
-/* Cascading referential integrity values     */
-/**********************************************/
-#define isc_dyn_foreign_key_update        205
-#define isc_dyn_foreign_key_delete        206
-#define isc_dyn_foreign_key_cascade       207
-#define isc_dyn_foreign_key_default       208
-#define isc_dyn_foreign_key_null          209
-#define isc_dyn_foreign_key_none          210
-
-/***********************/
-/* SQL role values     */
-/***********************/
-#define isc_dyn_def_sql_role              211
-#define isc_dyn_sql_role_name             212
-#define isc_dyn_grant_admin_options       213
-#define isc_dyn_del_sql_role              214
-/* 215 & 216 are used some lines above. */
-
-/**********************************************/
-/* Generators again                           */
-/**********************************************/
-
-#define isc_dyn_delete_generator          217
-
-// New for comments in objects.
-#define isc_dyn_mod_function              224
-#define isc_dyn_mod_filter                225
-#define isc_dyn_mod_generator             226
-#define isc_dyn_mod_sql_role              227
-#define isc_dyn_mod_charset               228
-#define isc_dyn_mod_collation             229
-#define isc_dyn_mod_prc_parameter         230
-
-/***********************/
-/* collation values    */
-/***********************/
-#define isc_dyn_def_collation						231
-#define isc_dyn_coll_for_charset					232
-#define isc_dyn_coll_from							233
-#define isc_dyn_coll_from_external					239
-#define isc_dyn_coll_attribute						234
-#define isc_dyn_coll_specific_attributes_charset	235
-#define isc_dyn_coll_specific_attributes			236
-#define isc_dyn_del_collation						237
-
-/******************************************/
-/* Mapping OS security objects to DB ones */
-/******************************************/
-#define isc_dyn_mapping								243
-#define isc_dyn_map_role							1
-#define isc_dyn_unmap_role							2
-#define isc_dyn_map_user							3
-#define isc_dyn_unmap_user							4
-#define isc_dyn_automap_role						5
-#define isc_dyn_autounmap_role						6
-
-/********************/
-/* Users control    */
-/********************/
-#define isc_dyn_user								244
-#define isc_dyn_user_add							1
-#define isc_dyn_user_mod							2
-#define isc_dyn_user_del							3
-#define isc_dyn_user_passwd							4
-#define isc_dyn_user_first							5
-#define isc_dyn_user_middle							6
-#define isc_dyn_user_last							7
-#define isc_dyn_user_admin							8
-#define isc_user_end								0
-
-/****************************/
-/* Last $dyn value assigned */
-/****************************/
-#define isc_dyn_last_dyn_value            247
-
 /******************************************/
 /* Array slice description language (SDL) */
 /******************************************/
@@ -916,26 +585,26 @@
 #define isc_sdl_tiny_integer              9
 #define isc_sdl_short_integer             10
 #define isc_sdl_long_integer              11
-#define isc_sdl_literal                   12
+//#define isc_sdl_literal                   12
 #define isc_sdl_add                       13
 #define isc_sdl_subtract                  14
 #define isc_sdl_multiply                  15
 #define isc_sdl_divide                    16
-#define isc_sdl_negate                    17
-#define isc_sdl_eql                       18
-#define isc_sdl_neq                       19
-#define isc_sdl_gtr                       20
-#define isc_sdl_geq                       21
-#define isc_sdl_lss                       22
-#define isc_sdl_leq                       23
-#define isc_sdl_and                       24
-#define isc_sdl_or                        25
-#define isc_sdl_not                       26
-#define isc_sdl_while                     27
-#define isc_sdl_assignment                28
-#define isc_sdl_label                     29
-#define isc_sdl_leave                     30
-#define isc_sdl_begin                     31
+#define isc_sdl_negate                    17 // only used in pretty.cpp; nobody generates it
+//#define isc_sdl_eql                       18
+//#define isc_sdl_neq                       19
+//#define isc_sdl_gtr                       20
+//#define isc_sdl_geq                       21
+//#define isc_sdl_lss                       22
+//#define isc_sdl_leq                       23
+//#define isc_sdl_and                       24
+//#define isc_sdl_or                        25
+//#define isc_sdl_not                       26
+//#define isc_sdl_while                     27
+//#define isc_sdl_assignment                28
+//#define isc_sdl_label                     29
+//#define isc_sdl_leave                     30
+#define isc_sdl_begin                     31 // only used in pretty.cpp; nobody generates it
 #define isc_sdl_end                       32
 #define isc_sdl_do3                       33
 #define isc_sdl_do2                       34
@@ -946,9 +615,9 @@
 /* International text interpretation values */
 /********************************************/
 
-#define isc_interp_eng_ascii              0
-#define isc_interp_jpn_sjis               5
-#define isc_interp_jpn_euc                6
+//#define isc_interp_eng_ascii              0
+//#define isc_interp_jpn_sjis               5
+//#define isc_interp_jpn_euc                6
 
 /*****************/
 /* Blob Subtypes */
@@ -973,17 +642,17 @@
 
 /* the range 20-30 is reserved for dBASE and Paradox types */
 
-#define isc_blob_formatted_memo           20
-#define isc_blob_paradox_ole              21
-#define isc_blob_graphic                  22
-#define isc_blob_dbase_ole                23
-#define isc_blob_typed_binary             24
+//#define isc_blob_formatted_memo           20
+//#define isc_blob_paradox_ole              21
+//#define isc_blob_graphic                  22
+//#define isc_blob_dbase_ole                23
+//#define isc_blob_typed_binary             24
 
 /* Deprecated definitions maintained for compatibility only */
 
-#define isc_info_db_SQL_dialect           62
-#define isc_dpb_SQL_dialect               63
-#define isc_dpb_set_db_SQL_dialect        65
+//#define isc_info_db_SQL_dialect           62
+//#define isc_dpb_SQL_dialect               63
+//#define isc_dpb_set_db_SQL_dialect        65
 
 /***********************************/
 /* Masks for fb_shutdown_callback  */
@@ -993,6 +662,7 @@
 #define fb_shut_preproviders			  2
 #define fb_shut_postproviders			  4
 #define fb_shut_finish					  8
+#define fb_shut_exit					 16
 
 /****************************************/
 /* Shutdown reasons, used by engine     */
@@ -1002,7 +672,7 @@
 #define fb_shutrsn_svc_stopped			  -1
 #define fb_shutrsn_no_connection		  -2
 #define fb_shutrsn_app_stopped			  -3
-#define fb_shutrsn_device_removed		  -4
+//#define fb_shutrsn_device_removed		  -4
 #define fb_shutrsn_signal				  -5
 #define fb_shutrsn_services				  -6
 #define fb_shutrsn_exit_called			  -7
@@ -1025,6 +695,9 @@
 #define fb_dbg_map_src2blr			2
 #define fb_dbg_map_varname			3
 #define fb_dbg_map_argument			4
+#define fb_dbg_subproc				5
+#define fb_dbg_subfunc				6
+#define fb_dbg_map_curname			7
 
 // sub code for fb_dbg_map_argument
 #define fb_dbg_arg_input			0
