@@ -146,7 +146,7 @@ private:
 };
 
 class Map;
-typedef Hash<Map, DEFAULT_HASH_SIZE, Map, DefaultKeyValue<Map>, Map> MapHash;
+typedef HashTable<Map, DEFAULT_HASH_SIZE, Map, DefaultKeyValue<Map>, Map> MapHash;
 
 class Map : public MapHash::Entry, public GlobalStorage
 {
@@ -581,40 +581,48 @@ class MappingIpc FB_FINAL : public Firebird::IpcObject
 
 public:
 	explicit MappingIpc(MemoryPool&)
-		: processId(getpid())
+		: processId(getpid()),
+		  cleanupSync(*getDefaultMemoryPool(), clearDelivery, THREAD_high)
 	{ }
 
 	~MappingIpc()
+	{
+		shutdown();
+	}
+
+	void shutdown()
 	{
 		if (!sharedMemory)
 			return;
 
-		Guard gShared(this);
-
-		MappingHeader* sMem = sharedMemory->getHeader();
-
-		startupSemaphore.tryEnter(5);
-		sMem->process[process].flags &= ~MappingHeader::FLAG_ACTIVE;
-		(void)  // Ignore errors in cleanup
-            sharedMemory->eventPost(&sMem->process[process].notifyEvent);
-		cleanupSemaphore.tryEnter(5);
-
-		// Ignore errors in cleanup
-		sharedMemory->eventFini(&sMem->process[process].notifyEvent);
-		sharedMemory->eventFini(&sMem->process[process].callbackEvent);
-
-		bool found = false;
-		for (unsigned n = 0; n < sMem->processes; ++n)
 		{
-			if (sMem->process[n].flags & MappingHeader::FLAG_ACTIVE)
-			{
-				found = true;
-				break;
-			}
-		}
+			Guard gShared(this);
+			MappingHeader* sMem = sharedMemory->getHeader();
 
-		if (!found)
-			sharedMemory->removeMapFile();
+			startupSemaphore.tryEnter(5);
+			sMem->process[process].flags &= ~MappingHeader::FLAG_ACTIVE;
+			(void)  // Ignore errors in cleanup
+				sharedMemory->eventPost(&sMem->process[process].notifyEvent);
+			cleanupSync.waitForCompletion();
+
+			// Ignore errors in cleanup
+			sharedMemory->eventFini(&sMem->process[process].notifyEvent);
+			sharedMemory->eventFini(&sMem->process[process].callbackEvent);
+
+			bool found = false;
+			for (unsigned n = 0; n < sMem->processes; ++n)
+			{
+				if (sMem->process[n].flags & MappingHeader::FLAG_ACTIVE)
+				{
+					found = true;
+					break;
+				}
+			}
+
+			if (!found)
+				sharedMemory->removeMapFile();
+		}
+		sharedMemory = NULL;
 	}
 
 	void clearMap(const char* dbName)
@@ -748,7 +756,7 @@ public:
 
 		try
 		{
-			Thread::start(clearDelivery, this, THREAD_high);
+			cleanupSync.run(this);
 		}
 		catch (const Exception&)
 		{
@@ -757,6 +765,12 @@ public:
 		}
 	}
 
+	void exceptionHandler(const Exception& ex, ThreadFinishSync<MappingIpc*>::ThreadRoutine*)
+	{
+		iscLogException("Fatal error in clearDeliveryThread", ex);
+		fb_utils::logAndDie("Fatal error in clearDeliveryThread");
+	}
+
 private:
 	void clearDeliveryThread()
 	{
@@ -794,13 +808,10 @@ private:
 			}
 			if (startup)
 				startupSemaphore.release();
-
-			cleanupSemaphore.release();
 		}
 		catch (const Exception& ex)
 		{
-			iscLogException("Fatal error in clearDeliveryThread", ex);
-			fb_utils::logAndDie("Fatal error in clearDeliveryThread");
+			exceptionHandler(ex, NULL);
 		}
 	}
 
@@ -855,11 +866,9 @@ private:
 		MappingIpc* const data;
 	};
 
-	static THREAD_ENTRY_DECLARE clearDelivery(THREAD_ENTRY_PARAM par)
+	static void clearDelivery(MappingIpc* mapping)
 	{
-		MappingIpc* m = (MappingIpc*)par;
-		m->clearDeliveryThread();
-		return 0;
+		mapping->clearDeliveryThread();
 	}
 
 	AutoPtr<SharedMemory<MappingHeader> > sharedMemory;
@@ -867,7 +876,7 @@ private:
 	const SLONG processId;
 	unsigned process;
 	Semaphore startupSemaphore;
-	Semaphore cleanupSemaphore;
+	ThreadFinishSync<MappingIpc*> cleanupSync;
 };
 
 GlobalPtr<MappingIpc, InstanceControl::PRIORITY_DELETE_FIRST> mappingIpc;
@@ -911,12 +920,14 @@ public:
 		embeddedSysdba.insertByte(isc_dpb_map_attach, TRUE);
 		embeddedSysdba.insertByte(isc_dpb_no_db_triggers, TRUE);
 
+		MAP_DEBUG(fprintf(stderr, "Attach %s\n", aliasDb));
 		IAttachment* att = prov->attachDatabase(&st, aliasDb,
 			embeddedSysdba.getBufferLength(), embeddedSysdba.getBuffer());
 
 		if (st->getState() & IStatus::STATE_ERRORS)
 		{
 			const ISC_STATUS* s = st->getErrors();
+			MAP_DEBUG(isc_print_status(s));
 			bool missing = fb_utils::containsErrorCode(s, isc_io_error);
 			down = fb_utils::containsErrorCode(s, isc_shutdown);
 			if (!(missing || down))
@@ -926,6 +937,7 @@ public:
 		}
 		else
 			reset(att);
+		MAP_DEBUG(fprintf(stderr, "Att=%p\n", att));
 
 		return down;
 	}
@@ -1338,4 +1350,9 @@ RecordBuffer* MappingList::getList(thread_db* tdbb, jrd_rel* relation)
 	return getData(relation);
 }
 
+void shutdownMappingIpc()
+{
+	mappingIpc->shutdown();
+}
+
 } // namespace Jrd
