@@ -28,6 +28,7 @@
 
 #include "firebird.h"
 #include <stdio.h>
+#include "../common/classes/Hash.h"
 #include "../jrd/jrd.h"
 #include "../jrd/lck.h"
 #include "gen/iberror.h"
@@ -59,7 +60,6 @@ static void bug_lck(const TEXT*);
 static bool compatible(const Lock*, const Lock*, USHORT);
 static void enqueue(thread_db*, CheckStatusWrapper*, Lock*, USHORT, SSHORT);
 static int external_ast(void*);
-static USHORT hash_func(const UCHAR*, USHORT);
 static void hash_allocate(Lock*);
 static Lock* hash_get_lock(Lock*, USHORT*, Lock***);
 static void hash_insert_lock(Lock*);
@@ -70,6 +70,7 @@ static void internal_dequeue(thread_db*, Lock*);
 static USHORT internal_downgrade(thread_db*, CheckStatusWrapper*, Lock*);
 static bool internal_enqueue(thread_db*, CheckStatusWrapper*, Lock*, USHORT, SSHORT, bool);
 static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type);
+static lck_owner_t get_owner_type(enum lck_t lock_type);
 
 #ifdef DEBUG_LCK
 namespace
@@ -508,6 +509,33 @@ static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 
 	SLONG handle = 0;
 
+	switch (get_owner_type(lock_type))
+	{
+	case LCK_OWNER_database:
+		handle = *LCK_OWNER_HANDLE_DBB(tdbb);
+		break;
+
+	case LCK_OWNER_attachment:
+		handle = *LCK_OWNER_HANDLE_ATT(tdbb);
+		break;
+
+	default:
+		bug_lck("Invalid lock owner type in get_owner_handle()");
+	}
+
+	if (!handle)
+	{
+		bug_lck("Invalid lock owner handle");
+	}
+
+	return handle;
+}
+
+
+static lck_owner_t get_owner_type(enum lck_t lock_type)
+{
+	lck_owner_t owner_type;
+
 	switch (lock_type)
 	{
 	case LCK_database:
@@ -519,7 +547,7 @@ static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 	case LCK_sweep:
 	case LCK_crypt:
 	case LCK_crypt_status:
-		handle = *LCK_OWNER_HANDLE_DBB(tdbb);
+		owner_type = LCK_OWNER_database;
 		break;
 
 	case LCK_attachment:
@@ -543,19 +571,14 @@ static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 	case LCK_btr_dont_gc:
 	case LCK_rel_gc:
 	case LCK_record_gc:
-		handle = *LCK_OWNER_HANDLE_ATT(tdbb);
+		owner_type = LCK_OWNER_attachment;
 		break;
 
 	default:
-		bug_lck("Invalid lock type in get_owner_handle()");
+		bug_lck("Invalid lock type in get_owner_type()");
 	}
 
-	if (!handle)
-	{
-		bug_lck("Invalid lock owner handle");
-	}
-
-	return handle;
+	return owner_type;
 }
 
 
@@ -954,38 +977,6 @@ static int external_ast(void* lock_void)
 }
 
 
-
-static USHORT hash_func(const UCHAR* value, USHORT length)
-{
-/**************************************
- *
- *	h a s h
- *
- **************************************
- *
- * Functional description
- *	Provide a repeatable hash value based
- *	on the passed key.
- *
- **************************************/
-
-	// Hash the value, preserving its distribution as much as possible
-
-	ULONG hash_value = 0;
-	UCHAR* p = 0;
-	const UCHAR* q = value;
-
-	for (USHORT l = 0; l < length; l++)
-	{
-		if (!(l & 3))
-			p = (UCHAR*) &hash_value;
-		*p++ += *q++;
-	}
-
-	return (USHORT) (hash_value % LOCK_HASH_SIZE);
-}
-
-
 static void hash_allocate(Lock* lock)
 {
 /**************************************
@@ -1036,7 +1027,8 @@ static Lock* hash_get_lock(Lock* lock, USHORT* hash_slot, Lock*** prior)
 	if (!att->att_compatibility_table)
 		hash_allocate(lock);
 
-	const USHORT hash_value = hash_func(lock->getKeyString(), lock->lck_length);
+	const USHORT hash_value = 
+		(USHORT) InternalHash::hash(lock->lck_length, lock->getKeyString(), LOCK_HASH_SIZE);
 
 	if (hash_slot)
 		*hash_slot = hash_value;
@@ -1473,20 +1465,20 @@ Lock::Lock(thread_db* tdbb, USHORT length, lck_t type, void* object, lock_ast_t
 
 void Lock::setLockAttachment(thread_db* tdbb, Jrd::Attachment* attachment)
 {
+	if (get_owner_type(lck_type) == LCK_OWNER_database)
+		return;
+
 	SET_TDBB(tdbb);
 	Database* dbb = tdbb->getDatabase();
 	fb_assert(dbb);
 	if (!dbb)
 		return;
 
-	Sync lckSync(&dbb->dbb_lck_sync, "setLockAttachment");
-	lckSync.lock(SYNC_EXCLUSIVE);
-
 	Attachment* att = lck_attachment ? lck_attachment->getHandle() : NULL;
 	if (att == attachment)
 		return;
 
-	// If lock has an attachment it must not be a part of linked list
+	// If lock has no attachment it must not be a part of linked list
 	fb_assert(!lck_attachment ? !lck_prior && !lck_next : true);
 
 	// Delist in old attachment
