@@ -27,32 +27,29 @@
 #include <stdio.h>
 
 #include "fb_blk.h"
+#include "consts_pub.h"
 
-#include "../jrd/jrd_pwd.h"
 #include "../jrd/svc_undoc.h"
-#include "../jrd/ThreadStart.h"
+#include "../common/ThreadStart.h"
 
 #include "../common/classes/semaphore.h"
 #include "../common/classes/array.h"
 #include "../common/classes/SafeArg.h"
-#include "../common/classes/PublicHandle.h"
 #include "../common/UtilSvc.h"
-#include "../burp/spit.h"
-
-#ifndef IO_BUFFER_SIZE
-#ifdef BUFSIZ
-const int SVC_IO_BUFFER_SIZE = (16 * (BUFSIZ));
-#else // BUFSIZ
-const int SVC_IO_BUFFER_SIZE = (16 * (1024));
-#endif // BUFSIZ
-#else // IO_BUFFER_SIZE
-const int SVC_IO_BUFFER_SIZE = (16 * (IO_BUFFER_SIZE));
-#endif // IO_BUFFER_SIZE
+#include "../jrd/EngineInterface.h"
+#include "../common/classes/Switches.h"
+#include "../common/classes/ClumpletReader.h"
+#include "../common/classes/RefMutex.h"
+#include "../burp/split/spit.h"
+#include "../jrd/status.h"
 
 // forward decl.
-struct serv_entry;
+
+namespace Jrd {
+	struct serv_entry;
+}
+
 namespace Firebird {
-	class ClumpletReader;
 	namespace Arg {
 		class StatusVector;
 	}
@@ -64,38 +61,38 @@ const ULONG SERVICE_VERSION			= 2;
 
 const int SVC_STDOUT_BUFFER_SIZE	= 1024;
 
-/* Flag of capabilities supported by the server */
-//const ULONG WAL_SUPPORT					= 0x1L;	// Write Ahead Log
-const ULONG MULTI_CLIENT_SUPPORT		= 0x2L;	/* SuperServer model (vs. multi-inet) */
-const ULONG REMOTE_HOP_SUPPORT			= 0x4L;	/* Server can connect to other server */
-//const ULONG NO_SVR_STATS_SUPPORT		= 0x8L;	// Does not support statistics
+// Flag of capabilities supported by the server
+//const ULONG WAL_SUPPORT				= 0x1L;		// Write Ahead Log
+const ULONG MULTI_CLIENT_SUPPORT		= 0x2L;		// SuperServer model (vs. multi-inet)
+const ULONG REMOTE_HOP_SUPPORT			= 0x4L;		// Server can connect to other server
+//const ULONG NO_SVR_STATS_SUPPORT		= 0x8L;		// Does not support statistics
 
-//const ULONG NO_DB_STATS_SUPPORT			= 0x10L;	// Does not support statistics
+//const ULONG NO_DB_STATS_SUPPORT		= 0x10L;	// Does not support statistics
 // Really the 16 bit LIBS here?
 //const ULONG LOCAL_ENGINE_SUPPORT		= 0x20L;	// The local 16 bit engine
-//const ULONG NO_FORCED_WRITE_SUPPORT		= 0x40L;	// Can not configure sync writes
-//const ULONG NO_SHUTDOWN_SUPPORT			= 0x80L;	// Can not shutdown/restart databases
-const ULONG NO_SERVER_SHUTDOWN_SUPPORT	= 0x100L;	/* Can not shutdown server */
-const ULONG SERVER_CONFIG_SUPPORT		= 0x200L;	/* Can configure server */
-const ULONG QUOTED_FILENAME_SUPPORT		= 0x400L;	/* Can pass quoted filenames in */
+//const ULONG NO_FORCED_WRITE_SUPPORT	= 0x40L;	// Can not configure sync writes
+//const ULONG NO_SHUTDOWN_SUPPORT		= 0x80L;	// Can not shutdown/restart databases
+const ULONG NO_SERVER_SHUTDOWN_SUPPORT	= 0x100L;	// Can not shutdown server
+//const ULONG SERVER_CONFIG_SUPPORT		= 0x200L;	// Can configure server
+const ULONG QUOTED_FILENAME_SUPPORT		= 0x400L;	// Can pass quoted filenames in
 
-/* Range definitions for service actions.  Any action outside of
-   this range is not supported */
+// Range definitions for service actions.  Any action outside of
+// this range is not supported
 const USHORT isc_action_min				= 1;
 const USHORT isc_action_max				= isc_action_svc_last;
 
-/* Range definitions for service actions.  Any action outside of
-   this range is not supported */
+// Range definitions for service actions.  Any action outside of
+// this range is not supported
 //define isc_info_min                  50
 //define isc_info_max                  67
 
-/* Bitmask values for the svc_flags variable */
-const int SVC_shutdown		= 0x1;
-const int SVC_timeout		= 0x2;
+// Bitmask values for the svc_flags variable
+//const int SVC_shutdown	= 0x1;
+//const int SVC_timeout		= 0x2;
 //const int SVC_forked		= 0x4;
 const int SVC_detached		= 0x8;
 const int SVC_finished		= 0x10;
-const int SVC_thd_running	= 0x20;
+//const int SVC_thd_running	= 0x20;
 const int SVC_evnt_fired	= 0x40;
 const int SVC_cmd_line		= 0x80;
 
@@ -104,7 +101,7 @@ class thread_db;
 class TraceManager;
 
 // Service manager
-class Service : public Firebird::UtilSvc, public TypedHandle<type_svc>, public Firebird::PublicHandle
+class Service : public Firebird::UtilSvc, public TypedHandle<type_svc>
 {
 public:		// utilities interface with service
 	// output to svc_stdout verbose info
@@ -112,21 +109,20 @@ public:		// utilities interface with service
 	// outpur error text
 	virtual void outputError(const char* text);
 	// output some data to service
-	virtual void outputData(const char* text);
+	virtual void outputData(const void* data, FB_SIZE_T len);
 	// printf() to svc_stdout
     virtual void printf(bool err, const SCHAR* format, ...);
 	// returns true - it's service :)
 	virtual bool isService();
 	// client thread started
 	virtual void started();
-	// client thread finished
-	virtual void finish();
 	// put various info items in info buffer
     virtual void putLine(char tag, const char* val);
     virtual void putSLong(char tag, SLONG val);
+    virtual void putSInt64(char tag, SINT64 val);
 	virtual void putChar(char tag, char val);
 	// put raw bytes to svc_stdout
-	virtual void putBytes(const UCHAR*, size_t);
+	virtual void putBytes(const UCHAR*, FB_SIZE_T);
 	// get raw bytes from svc_stdin
 	virtual ULONG getBytes(UCHAR*, ULONG);
 	// append status_vector to service's status
@@ -136,13 +132,15 @@ public:		// utilities interface with service
 	// no-op for services
 	virtual void hidePasswd(ArgvType&, int);
 	// return service status
-    virtual const ISC_STATUS* getStatus();
+    virtual const FbStatusVector* getStatus();
 	// reset service status
 	virtual void initStatus();
 	// no-op for services
 	virtual void checkService();
-	// add address path (taken from spb) to dpb if present
-	virtual void getAddressPath(Firebird::ClumpletWriter& dpb);
+	// add address path and utf8 flag (taken from spb) to dpb if present
+	virtual void fillDpb(Firebird::ClumpletWriter& dpb);
+	// encoding for string parameters passed to utility
+	virtual bool utf8FileNames();
 
 	virtual TraceManager* getTraceManager()
 	{
@@ -155,9 +153,13 @@ public:		// utilities interface with service
 			|| checkForShutdown();
 	}
 
+	// Get authentication block if present
+	virtual unsigned int getAuthBlock(const unsigned char** bytes);
+
 public:		// external interface with service
 	// Attach - service ctor
-	Service(const TEXT* service_name, USHORT spb_length, const UCHAR* spb_data);
+	Service(const TEXT* service_name, USHORT spb_length, const UCHAR* spb_data,
+		Firebird::ICryptKeyCallback* crypt_callback);
 	// Start service thread
 	void start(USHORT spb_length, const UCHAR* spb_data);
 	// Query service state (v. 1 & 2)
@@ -174,7 +176,7 @@ public:		// external interface with service
 	}
 
 	// Firebird log reader
-	static THREAD_ENTRY_DECLARE readFbLog(THREAD_ENTRY_PARAM arg);
+	static int readFbLog(Firebird::UtilSvc* uSvc);
 	// Shuts all service threads (should be called after databases shutdown)
 	static void shutdownServices();
 
@@ -186,9 +188,6 @@ public:		// external interface with service
 
 	const Firebird::string&	getUserName() const
 	{
-		if (svc_username.empty())
-			return svc_trusted_login;
-
 		return svc_username;
 	}
 
@@ -196,27 +195,31 @@ public:		// external interface with service
 	const Firebird::string&	getRemoteAddress() const	{ return svc_remote_address; }
 	const Firebird::string&	getRemoteProcess() const	{ return svc_remote_process; }
 	int	getRemotePID() const { return svc_remote_pid; }
+	const Firebird::PathName& getExpectedDb() const		{ return svc_expected_db; }
+	Firebird::ICryptKeyCallback* getCryptCallback()		{ return svc_crypt_callback; }
 
 private:
 	// Service must have private destructor, called from finish
 	// when both (server and client) threads are finished
 	~Service();
 	// Find current service in global services list
-	bool	locateInAllServices(size_t* posPtr = NULL);
+	bool	locateInAllServices(FB_SIZE_T* posPtr = NULL);
 	// Detach self from global services list
 	void	removeFromAllServices();
 	// The only service, implemented internally
 	void	readFbLog();
 	// Create argv, argc and svc_parsed_sw
 	void	parseSwitches();
+	// Check does this action need arg or not
+	static bool actionNeedsArg(UCHAR action);
 	// Put data into stdout buffer
 	void	enqueue(const UCHAR* s, ULONG len);
 	// true if there is no data in stdout buffer
-	bool	empty() const;
+	bool	empty(ULONG head) const;
 	// true if no more space in stdout buffer
 	bool	full() const;
 	// start service thread
-	void	start(ThreadEntryPoint* service_thread);
+	void	start(const serv_entry* service_run);
 	// Set the flag (either SVC_finished for the main service thread or SVC_detached for the client thread).
 	// If both main thread and client thread are completed that is main thread is finished and
 	// client is detached then free memory used by service.
@@ -229,44 +232,59 @@ private:
 	// Returns number of bytes service wants more
 	ULONG	put(const UCHAR* buffer, ULONG length);
 	// Copies argument value to status vector
-	void put_status_arg(ISC_STATUS*& status, const MsgFormat::safe_cell& value);
+	void put_status_arg(Firebird::Arg::StatusVector& status, const MsgFormat::safe_cell& value);
 
 	// Increment circular buffer pointer
 	static ULONG		add_one(ULONG i);
 	static ULONG		add_val(ULONG i, ULONG val);
 	// Convert spb flags to utility switches
-	static void			conv_switches(Firebird::ClumpletReader& spb, Firebird::string& switches);
+#ifndef DEV_BUILD
+	static
+#endif
+	void				conv_switches(Firebird::ClumpletReader& spb, Firebird::string& switches);
 	// Find spb switch in switch table
-	static const TEXT*	find_switch(int in_spb_sw, const in_sw_tab_t* table);
+	static const TEXT*	find_switch(int in_spb_sw, const Switches::in_sw_tab_t* table, bool bitmask);
 	// Loop through the appropriate switch table looking for the text for the given command switch
-	static bool			process_switches(Firebird::ClumpletReader& spb, Firebird::string& switches);
+#ifndef DEV_BUILD
+	static
+#endif
+	bool				process_switches(Firebird::ClumpletReader& spb, Firebird::string& switches);
 	// Get bitmask from within spb buffer, find corresponding switches within specified table,
 	// add them to the command line
-	static bool get_action_svc_bitmask(const Firebird::ClumpletReader& spb, const in_sw_tab_t* table,
+	static bool get_action_svc_bitmask(const Firebird::ClumpletReader& spb,
+									   const Switches::in_sw_tab_t* table,
 									   Firebird::string& sw);
 	// Get string from within spb buffer, add it to the command line
 	static void get_action_svc_string(const Firebird::ClumpletReader& spb, Firebird::string& sw);
+	// Get string from within spb buffer, insert it at given position into command line
+	static void get_action_svc_string_pos(const Firebird::ClumpletReader& spb, Firebird::string& switches,
+										  Firebird::string::size_type p = Firebird::string::npos);
 	// Get integer from within spb buffer, add it to the command line
-	static void get_action_svc_data(const Firebird::ClumpletReader& spb, Firebird::string& sw);
+	static void get_action_svc_data(const Firebird::ClumpletReader& spb, Firebird::string& sw, bool bigint);
 	// Get parameter from within spb buffer, find corresponding switch within specified table,
 	// add it to the command line
-	static bool get_action_svc_parameter(UCHAR tag, const in_sw_tab_t* table, Firebird::string&);
+	static bool get_action_svc_parameter(UCHAR tag, const Switches::in_sw_tab_t* table,
+										 Firebird::string&);
 	// Create 'SYSDBA needed' error in status vector
 	static void need_admin_privs(Firebird::Arg::StatusVector& status, const char* message);
 	// Does info buffer have enough space for SLONG?
 	static bool ck_space_for_numeric(UCHAR*& info, const UCHAR* const end);
 	// Make status vector permamnent, if one present in worker thread's space
 	void makePermanentStatusVector() throw();
+	// Read SPB on attach
+	void getOptions(Firebird::ClumpletReader&);
+	// Invoke appropriate service thread entry and finalize it correctly
+	static THREAD_ENTRY_DECLARE run(THREAD_ENTRY_PARAM arg);
 
 private:
-	ISC_STATUS_ARRAY svc_status;		// status vector for running service
-	Firebird::string svc_parsed_sw;		// Here point elements of argv
+	FbLocalStatus svc_status;						// status vector for running service
+	Firebird::string svc_parsed_sw;					// Here point elements of argv
 	ULONG	svc_stdout_head;
 	ULONG	svc_stdout_tail;
 	UCHAR	svc_stdout[SVC_STDOUT_BUFFER_SIZE];		// output from service
 	Firebird::Semaphore	svcStart;
-	const serv_entry*	svc_service;			// attached service's entry
-	const serv_entry*	svc_service_run;		// running service's entry
+	const serv_entry*	svc_service;				// attached service's entry
+	const serv_entry*	svc_service_run;			// running service's entry
 	Firebird::Array<UCHAR> svc_resp_alloc;
 	UCHAR*	svc_resp_buf;
 	const UCHAR*	svc_resp_ptr;
@@ -276,17 +294,21 @@ private:
 	USHORT	svc_user_flag;
 	USHORT	svc_spb_version;
 	bool	svc_do_shutdown;
+	bool	svc_shutdown_in_progress;
+	bool	svc_timeout;
 	char	svc_arg_conv[MsgFormat::SAFEARG_MAX_ARG * 2];
 	char*	svc_arg_ptr;
 
 	Firebird::string	svc_username;
-	Firebird::string	svc_enc_password;
-	Firebird::string	svc_trusted_login;
+	Firebird::string	svc_sql_role;
+	Firebird::AuthReader::AuthBlock	svc_auth_block;
+	Firebird::PathName	svc_expected_db;
 	bool                svc_trusted_role;
-	bool				svc_uses_security_database;
+	bool				svc_utf8;
 	Firebird::string	svc_switches;	// Full set of switches
 	Firebird::string	svc_perm_sw;	// Switches, taken from services table and/or passed using spb_command_line
 	Firebird::string	svc_address_path;
+	Firebird::string	svc_command_line;
 
 	Firebird::string	svc_network_protocol;
 	Firebird::string	svc_remote_address;
@@ -294,26 +316,63 @@ private:
 	SLONG				svc_remote_pid;
 
 	TraceManager*		svc_trace_manager;
+	Firebird::ICryptKeyCallback* svc_crypt_callback;
 
 public:
-	struct StatusStringsHelper
-	{
-		FB_THREAD_ID workerThread;
-		Firebird::Mutex mtx;
-	};
-
 	Firebird::Semaphore	svc_detach_sem;
 
-private:
-	StatusStringsHelper	svc_thread_strings;
+	class SvcMutex : public Firebird::RefMutex
+	{
+	public:
+		explicit SvcMutex(Service* svc)
+			: link(svc)
+		{ }
 
+		Service* link;
+	};
+
+	Firebird::RefPtr<SvcMutex> svc_existence;
+
+private:
 	Firebird::Semaphore svc_sem_empty, svc_sem_full;
 
+	class Validate
+	{
+	public:
+		explicit Validate(Service* svc);
+		Firebird::MutexEnsureUnlock sharedGuard;
+	};
+
+	class SafeMutexLock : private Validate
+	{
+	public:
+		SafeMutexLock(Service* svc, const char* f);
+		bool lock();
+
+	protected:
+		Firebird::RefPtr<SvcMutex> existenceMutex;
+		const char* from;
+	};
+
+	friend class SafeMutexLock;
+
 	//Service existence guard
-	class ExistenceGuard : public Firebird::PublicHandleHolder
+	class ExistenceGuard : private SafeMutexLock
 	{
 	public:
 		explicit ExistenceGuard(Service* svc, const char* from);
+		~ExistenceGuard();
+	};
+
+	//Service unlock guard
+	class UnlockGuard : private SafeMutexLock
+	{
+	public:
+		explicit UnlockGuard(Service* svc, const char* from);
+		bool enter();
+		~UnlockGuard();
+	private:
+		bool locked, doLock;
 	};
 
 	// Data pipe from client to service
@@ -332,6 +391,10 @@ private:
 	// Size of data, placed into svc_stdin_buffer (set in put)
 	ULONG svc_stdin_user_size;
 	static const ULONG PRELOAD_BUFFER_SIZE = SVC_IO_BUFFER_SIZE;
+
+#ifdef DEV_BUILD
+	bool svc_debug;
+#endif
 };
 
 } //namespace Jrd
